program lndbdd;
uses Windows;


///////////////////////////////////\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
/////////////////////////////// Consts \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
///////////////////////////////////\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
const
 VerData = 'Lnd Block Data Dumper v1.00';

 OpenLNDTitle: pchar = 'Open LND file';
 LNDFilter: packed array[1..30] of char = (
  'B', '&', 'W', ' ', 'l', 'a', 'n', 'd', ' ', 'f', 'i', 'l', 'e', 's', ' ', '(', '*', '.', 'l', 'n', 'd', ')', #0, '*', '.', 'L', 'N', 'D', #0,
  #0);
 SaveRawTitle: pchar = 'Save raw image file';
 RawFilter: packed array[1..31] of char = (
  'R', 'a', 'w', ' ', 'i', 'm', 'a', 'g', 'e', ' ', 'f', 'i', 'l', 'e', 's', ' ', '(', '*', '.', 'r', 'a', 'w', ')', #0, '*', '.', 'R', 'A', 'W', #0,
  #0);
 ContinueTitle: pchar = 'Continue working?';
 ContinueMsg: pchar = 'Do you wish to dump another type of LND data now?';

 FlagMask: array[2..3] of byte = (
  $0F, // COUNTRY
  $80 // SPLIT
 ); // Leaving out WATER and COAST, since Lndsert & Lndromat fill those in automagically based on altitude.


///////////////////////////////////\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
////////////////////////////// Typedefs \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
///////////////////////////////////\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
type
///////////////////////// LND crap \\\\\\\\\\\\\\\\\\\\\\\\\
//TLndFile = imaginary record
// Header: TLndHeader;
// LoResTexture: array[1..Header.NumLoResTextures] of TLndLoResTex;
// Block: array[1..Header.NumBlocks] of TLndBlock;
// Country: array[1..Header.NumCountries] of TLndCountry;
// Material: array[1..Header.NumMaterials] of TLndMaterial;
// NoiseMap: array[0..255,0..255] of byte;
// BumpMap: array[0..255,0..255] of byte;
//end;

 TLndHeader = packed record
  NumBlocks: longint;
  BlockIndex: array[0..31,0..31] of byte; // [Y,X]
  NumMaterials: longint;
  NumCountries: longint;
  BlockSize: longint;    // sizeof(TLndBlock)    =   2520
  MaterialSize: longint; // sizeof(TLndMaterial) = 131074
  CountrySize: longint;  // sizeof(TLndCountry)  =   3076
  NumLoResTextures: longword;
 end;

 TLndLoResTex = packed record
  Texture: pointer; // ignore
  Material: pointer; // ignore
  NumSubTextures: longint;
  ID: longint;
  Size: longint;
  // (Size - 4) bytes of DirectDraw texture crap
 end;

 TLndCell = packed record
  Color: array[1..4] of byte; // or something
  Altitude: byte;
  SaveColor: byte;
  Flags: array[1..2] of byte;
 end;
 TLndBlock = packed record
  Cell: array[0..16,0..16] of TLndCell;
  Index: longint;
  MapY: single; // scripting X coord = BlockX * 160
  MapX: single; // scripting Y coord = BlockY * 160
  BlockY: longint; // X coord in Header.BlockIndex
  BlockX: longint; // Y coord in Header.BlockIndex

  Dummy: array[1..188] of byte;
 end;


///////////////////////////////////\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
//////////////////////////////// Vars \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
///////////////////////////////////\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
var
 LND, DumpFile: HANDLE;
 LNDOFN, DumpOFN: OPENFILENAME;
 FNbuf: array[1..320] of char;
 Header: TLndHeader;
 LoResTexture: TLndLoResTex;
 Block: TLndBlock;
 BigMap: array[0..511,0..511] of byte;
 P: pointer;
 X, Y, Z, BytesRead, BytesWritten, Anchor: longword;
 i, j: word;
 DumpDataType: byte;


///////////////////////////////////\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
//////////////////////////////// Code \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
///////////////////////////////////\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
begin
 writeln(VerData);
 writeln;

 // Skip this if command line gives a filename?
 LNDOFN.lStructSize := sizeof(LNDOFN);
 LNDOFN.lpstrFilter := @LNDFilter;
 LNDOFN.lpstrFile := @FNbuf;
 LNDOFN.nMaxFile := sizeof(FNbuf);
 LNDOFN.lpstrTitle := OpenLNDTitle;
 LNDOFN.Flags := OFN_EXPLORER + OFN_FILEMUSTEXIST + OFN_HIDEREADONLY;
 if GetOpenFileName(@LNDOFN) then
  writeln('Loading ', FNbuf, '...')
 else // We have no LND file to edit.
  exit;

 LND := CreateFile(@FNbuf, (GENERIC_READ + GENERIC_WRITE), FILE_SHARE_READ, nil, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
 if (LND = INVALID_HANDLE_VALUE) then
  begin
   writeln('Error ', GetLastError, ' opening ', FNbuf, '!');
   exit;
  end;

////////////////////////// Header \\\\\\\\\\\\\\\\\\\\\\\\\\
 if not ReadFile(LND, Header, SizeOf(Header), BytesRead, nil) then
  begin
   writeln('Error ', GetLastError, ' reading LND header!');
   exit;
  end;

 writeln('Header info:');
 writeln(' NumBlocks        ', Header.NumBlocks);
 writeln(' NumMaterials     ', Header.NumMaterials);
 writeln(' NumCountries     ', Header.NumCountries);
 writeln(' BlockSize        ', Header.BlockSize);
 writeln(' MaterialSize     ', Header.MaterialSize);
 writeln(' CountrySize      ', Header.CountrySize);
 writeln(' NumLoResTextures ', Header.NumLoResTextures);
 writeln;

///////////////////// Low-Res Textures \\\\\\\\\\\\\\\\\\\\\
 for i := 1 to Header.NumLoResTextures do
  begin // skip low-res textures
   ReadFile(LND, LoResTexture, SizeOf(LoResTexture), BytesRead, nil);
   SetFilePointer(LND, (LoResTexture.Size - 4), nil, FILE_CURRENT);
  end;

///////////////////// Blocks and Cells \\\\\\\\\\\\\\\\\\\\\
 Anchor := SetFilePointer(LND, 0, nil, FILE_CURRENT);
 repeat
  SetFilePointer(LND, Anchor, nil, FILE_BEGIN);
  writeln('Select cell data to dump:');
  writeln;
  writeln(' 1. Altitude');
  writeln(' 2. Country');
  writeln(' 3. Split flag');
  writeln;
  writeln(' 0. Quit');
  writeln;
  repeat
   write('> '); readln(DumpDataType);
   writeln;
  until (DumpDataType >= 0) and (DumpDataType < 4);
  if (DumpDataType = 0) then
   begin
    CloseHandle(LND);
    exit;
   end;
 
  FNbuf := 'BigMap.raw';
  DumpOFN.lStructSize := sizeof(DumpOFN);
  DumpOFN.lpstrFilter := @RawFilter;
  DumpOFN.lpstrFile := @FNbuf;
  DumpOFN.nMaxFile := sizeof(FNbuf);
  DumpOFN.lpstrTitle := SaveRawTitle;
  DumpOFN.Flags := OFN_EXPLORER + OFN_OVERWRITEPROMPT;
  if GetSaveFileName(@DumpOFN) then
   begin
    DumpFile := CreateFile(@FNbuf, (GENERIC_READ + GENERIC_WRITE), FILE_SHARE_READ, nil, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, 0);
 
    for i := 2 to Header.NumBlocks do // why the hell...?
     begin // dump block cell data
      ReadFile(LND, Block, SizeOf(Block), BytesRead, nil);
      writeln('Dumping block #', Block.Index);
 
      if DumpDataType = 1 then
       for Y := 0 to 16 do
        for X := 0 to 16 do
         BigMap[(Block.BlockY * 16 + Y), (Block.BlockX * 16 + X)] := Block.Cell[Y,X].Altitude
      else if DumpDataType < 7 then
       for Y := 0 to 15 do
        for X := 0 to 15 do
         BigMap[(Block.BlockY * 16 + Y), (Block.BlockX * 16 + X)] := Block.Cell[Y,X].Flags[1] and FlagMask[DumpDataType];
 
     end;
 
    WriteFile(DumpFile, BigMap, SizeOf(BigMap), BytesWritten, nil);
    CloseHandle(DumpFile);
   end;
 until (MessageBox(0, ContinueMsg, ContinueTitle, MB_YESNO) = IDNO);

 CloseHandle(LND);
end.
